<?php
/*_______________________________________________
|                                                |
|    ©2012 Element Technologie - openElement     |
|________________________________________________|

	Using different SQL engines to handle DB and process packs of SQL requests
	See comments in the code
*/

$dir_dbrequestphp = dirname(__FILE__); // folder of this script
//$dir_dbrequestphp_to_tools = dirname(dirname($dir_dbrequestphp)); // corresponds to ../../
$dir_dbrequestphp_to_db = dirname($dir_dbrequestphp); // corresponds to ../
require_once $dir_dbrequestphp."/OEDynUtils.php";
require_once $dir_dbrequestphp."/OEDataContainers.php";

require_once $dir_dbrequestphp_to_db."/sqllogin.php"; // OE-specified login parameters

//_________________________________
// Base class for sql handlers ////////////////////////////////////
class OESQLBase {

	// Connection:
	public $sqlConn; 			// database connection, can be of different type according to chosen engine
	//protected $sqlConnParams;	// connection parameters, like db name, username/password - depends on chose engine
	
	// Errors:
	public $connectError; 		// connection error
	public $lastQueryError; 	// last SQL query error
	
	private static $sqlInstance; // instance of one of descendant classes depending on chosen engine
	
	// Public functions (do not override directly) /////////////////////
	
	// Return - create if not yet done - an instance of SQL handler according to chosen mode/engine
	public static function SQLInstance() {
		if (self::$sqlInstance) return self::$sqlInstance;
		$type = 'MySQL';
		if (OESQLBase::isLocalDB()) {
			global $oeSQLEngine_Local; if (isset($oeSQLEngine_Local)) $type = $oeSQLEngine_Local;
		} else {
			global $oeSQLEngine; if (isset($oeSQLEngine)) $type = $oeSQLEngine;
		}
		if (OESQLBase::isLocalSQLite()) {
			$type = 'SQLite';
		}
		echo '  = = Connecting to '.$type.'<br>';
		
		switch ($type) {
			case 'SQLite': 	
				self::$sqlInstance = new OESQLite(); break;
			default:
				self::$sqlInstance = new OEMySQL(); break;
		}
		//var_dump(self::$sqlInstance);
		return self::$sqlInstance;
	}	
	
	// Connect to database
	public function connect($params = null) {
		if (!$params) $params = $this->_getConnectParams();
		//var_dump($params);
		if (!$params) return false;
		return $this->_connect($params);
	}
	
	// Run a simple SQL query
	public function sql($query) {
		if (!$this->_checkInit($query)) return false;
		return $this->_sql($query);		
	}
	
	public function runSQLPack($sqlPack) {
		if (!$this->_checkInit($sqlPack)) return false;
		//var_dump($sqlPack);
		return $this->_runSQLPack($sqlPack);
	}
	
	// Secure value used in queries from SQL injections
	public function secureSQLValue($val, $forceNumeric = false) {
		if (($val === null || $val === '') && $forceNumeric) return null; //!!improve
		$len = strlen($val); if ($len < 1) return '';
		$val2 = $val;
		$val2 = $this->_sqlEscapeString($val2);
		if ($forceNumeric) { // if numeric mode demanded (to avoid injections like "1 or 1=1" => "..WHERE users.password=1 or 1=1.." and other):
			if (strpos($val2, '.') !== false) $val2 = (float)$val2; else $val2 = (int)$val2;
		}
		return $val2;
	}
	
	// ____________________
	// Non-overridable ///////////////////////////////
	
	public static function isLocalDB() { //echo __FILE__;
		return (strpos(__FILE__, ':') === 1);
	}

	public static function isLocalSQLite() { 
		return OEDynUtils::isLocalWebServer();
	}
	
	protected function _checkInit($val = true) {
		if (!$val) return false;
		if (!$this->sqlConn && !$this->connectError) $this->connect();
		//var_dump($this);
		if (!$this->sqlConn) return false;
		//echo 'Connected Ok ';var_dump($this->sqlConn);
		return true;
	}

	
	// ___________________________
	// Optionally overridable ///////////////////////////////////
	
	// run a pack of SQL requests generated by OE
	protected function _runSQLPack($sql) {
		/*	$sql is  an object containing requests, values for placeholders, and special instructions
			Format: {"Requests":[{"Type":1,"Code":"SELECT *","Values":["1","\"2\""],"ValueTypes":"is"},....]}
			or simply an array: [{"Type":1,"Code":"SELECT *","Values":["1","\"2\""],"ValueTypes":"is"},....]
		*/

		$arResultContainers = array();

		// if still in json serialisation, deserialise:
		if (!is_object($sql) && !is_array($sql)) {
			$RContainer = OEDynUtils::decodeJSON($sql);
		} else
			$RContainer = $sql;
		//var_dump($RContainer);//var_dump($RContainer->Requests[0]->Values);

		if ($RContainer && is_object($RContainer) && $RContainer->Requests) $RContainer = $RContainer->Requests;
		if (!$RContainer || count($RContainer)<1) return null;

		$rLastResult = null; // last mysqli request result
		$cntrSQL = 0; // counter of sql requests
		$InsertedIDs = array(); // store last insert id (last written autoinc value) for an sql request[i] (i=$cntrSQL)
		$aiPref = "__INAI"; // probably not used atm // keep same as DBTools.vb value!

		foreach ($RContainer as $rq) {
			$rType = $rq->Type;
			$rCode = $rq->Code;
			$rValues =& $rq->Values;
			
			// fix SQL query format:
			$arExtraQueries = array();
			$rCode = $this->_fixQueryFormat($rCode, $arExtraQueries);
			
			// Put values into SQL:
			
			// probably not used atm // replace foreign key "targets" with respective autoinc values
			/*if ($rValues && count($rValues)) {
				$rValCount = count($rValues);
				for ($i=0; $i<$rValCount; $i++) {
					if (strpos($rValues[$i], $aiPref) === 0) {
						$indTargetTable = substr($rValues[$i], strlen($aiPref));
						$rValues[$i] = $InsertedIDs[$indTargetTable];
					}
				}
			}*/
			
			// replace "?" with values
			$rCodeWithVals = ""; // SQL with "?" replaced by values
			$cntrVals = 0;
			if (strpos($rCode, '?') !== false && $rValues) {
				$rCodeSplit = explode('?', $rCode);
				$splitCount = count($rCodeSplit);
				for ($i=1; $i<$splitCount; $i++) { // add value to beginning of each split item except first - this will replace ? with value:
					$rCodeSplit[$i] = $rValues[$cntrVals].$rCodeSplit[$i]; // replace "?" with value
					$cntrVals++; // pass to next placeholder and value
				}
				$rCodeWithVals = implode($rCodeSplit);
			} else 
				$rCodeWithVals = $rCode;
			//echo "Request: $rCodeWithVals <br>"; // $rType $rCode $rValues <br> 
			
			if ($rCodeWithVals && strlen($rCodeWithVals) > 6) {
				//echo $rCodeWithVals."<br/>";
				
				// run SQL query:
				$sqlResult = $this->sql($rCodeWithVals);
				//var_dump($sqlResult);
				if ($sqlResult === false || $sqlResult === null) {
					continue;
				}
				
				// only for certain query types - store results in data container:
				$queryCodeStart = strtolower(substr($rCodeWithVals, 0, 6));
				
				if ($queryCodeStart === "select") { // store results of Select
				
					$rContainer = $this->_createSQLResultContainer();
					$rContainer->addAllRows($sqlResult);
					array_push($arResultContainers, $rContainer);
					//echo $rContainer->toJSON(); // ."<br>" // serialise result of every request
					//echo ' SELECT result:';var_dump($rContainer);
				
				} else 
				if ($queryCodeStart === "insert") { // store autoinc value for insert requests
					/*$InsertedIDs[$cntrSQL] = $this->_sqlInsertID();*///echo 'Inserted row '.$this->_sqlInsertID().'<br>';
					
					// in result container, add the source "last inserted primary key value":
					$rContainer = $this->_createSQLResultContainer();
					$rContainer->Values['_NewRowPrimKey'] = $this->_sqlInsertID(); // keep name '_NewRowPrimKey' in sync with vb!
					//echo 'INSERT new row result:'.$rContainer->Values['_NewRowPrimKey'];var_dump($rContainer->Values);
					array_push($arResultContainers, $rContainer);
				}
				
				if (!empty($arExtraQueries)) {
					foreach($arExtraQueries as $extraSQL) {
						$this->sql($extraSQL); // run extra queries in simple mode, ex create indexes after creating table in sqlite mode
					}
				}
				$cntrSQL++;
			}		
				
		} // for each request in the pack
		
		if (count($arResultContainers) < 1) 
			return null; 
		else 
			return $arResultContainers;	
		
	}

	
	// ___________________________
	// Overridable functions ///////////////////////////////////
	
	// Constructor
	function __construct() {		
	}
	
	protected function _getConnectParams() {
		return null;
	}
	
	// Connect to database
	protected function _connect($params) {
		return false;
	}
	
	// Run either a simple SQL query
	protected function _sql($query) {
		return null;
	}
	
	// auto primary key value for last insert operation
	protected function _sqlInsertID() {
		return 0;
	}
	
	protected function _createSQLResultContainer() {
		return new OESQLResult();
	}

	// normally called BEFORE applying values to placeholders, to correct formats, ex for SQLite ` => " and " => '
	protected function _fixQueryFormat($query, &$arExtraQueries) {
		return $query;
	}
	
	protected function _sqlEscapeString($val) {
		return $val;
	}
	
} 



//__________________________
// MySQL engine wrapper ////////////////////////////////////
class OEMySQL extends OESQLBase {


	// Overridable functions ///////////////////////////////////
	
	protected function _getConnectParams() {
		if (OESQLBase::isLocalDB()) {
			global $oeLoginMySQL_Local; if (!isset($oeLoginMySQL_Local)) return null;
			return $oeLoginMySQL_Local;
		} else {
			global $oeLoginMySQL; if (!isset($oeLoginMySQL)) return null;
			return $oeLoginMySQL;
		}
		return null;
	}
	
	// Connect to database
	protected function _connect($params) {
		if (!isset($params['server']) || !isset($params['user'])) return false;
		
		$mysqli_con = mysqli_connect($params['server'], $params['user'], $params['pw'], $params['dbname']);
		if (!$mysqli_con) {
			$this->connectError = mysqli_error($mysqli_con); 
			return false;  
		}
		
		$r_db = mysqli_select_db($mysqli_con, $params['dbname']); //!!check if necessary
		if (!$r_db) {
			$this->connectError = 'Error: no database "'.$params['dbname'].'" found';
			return false;
		}
		
		mysqli_set_charset($mysqli_con, 'utf8');
		
		$this->sqlConn = $mysqli_con;
	}
	
	// Run a simple SQL query
	protected function _sql($query) {
		if (!$query || !$this->sqlConn) return false;
		$mysqli_con = $this->sqlConn;
		$r = null; 
		echo '<span style="font-size:7px;">'.$query.'</span><br>';
		
		$msqliR = $mysqli_con->query($query, MYSQLI_STORE_RESULT);
		if (!$msqliR) {
			$this->lastQueryError = mysqli_error($mysqli_con);
			echo $this->lastQueryError;
			return false;
		} else if ($msqliR === true) {
			return true; // no-result queries, like INSERT
		}
		
		
		$r = $msqliR; //mysqli_store_result($mysqli_con);
		//var_dump($r);
		
		if (false) { // debug
			$query = str_ireplace("Select ", "<br>SELECT ", $query);
			$query = str_ireplace(" From ", 	"<br>FROM ", $query);
			$query = str_ireplace(" Where ", 	"<br>WHERE ", $query);
			;var_dump($r);echo ": ";var_dump($query);echo "<br>_ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _<br>";
		}
		return $r;
	}
	
	// auto primary key value for last insert operation
	protected function _sqlInsertID() {
		if (!$this->sqlConn) return 0;
		return $this->sqlConn->insert_id;
	}
	
	
	protected function _sqlEscapeString($val) {
		if (!$this->_checkInit()) return $val;
		return mysqli_real_escape_string($this->sqlConn, $val);
	}
	
	
} // end class OEMySQL



//__________________________
// SQLite engine wrapper ////////////////////////////////////
class OESQLite extends OESQLBase {

	private $sqliteOk;
	private $sqliteVers;
	
	// Overridable functions ///////////////////////////////////
	
	// Constructor
	function __construct() {
		$this->sqliteOk = false;
		$this->sqliteVers = 0;
		if (class_exists('SQLite3')) {
			echo ' ..SQLite3 class exists..<br/>';
			$this->sqliteVers = 3;
			$this->sqliteOk = true;
		} else {
			$this->sqliteOk = function_exists('sqlite_escape_string'); //!!improve
			$this->sqliteVers = 2;
			echo ' ..No SQLite3.. ';
			//phpinfo();
		}
		//echo "<br/>SQLite activated, version {$this->sqliteVers} <br/>";
	}
	
	protected function _getConnectParams() {
		if (OESQLBase::isLocalDB()) {
			global $oeLoginSQLite_Local; if (!isset($oeLoginSQLite_Local)) return null;
			//echo 'Connect params:';var_dump($oeLoginSQLite_Local);
			return $oeLoginSQLite_Local;
		} else {
			global $oeLoginSQLite; if (!isset($oeLoginSQLite)) return null;
			//echo 'Connect params:';var_dump($oeLoginSQLite);
			return $oeLoginSQLite;
		}
		return null;
	}
	
	// Connect to database
	protected function _connect($params) {
		//echo 'Connecting start:';var_dump($this);
		if (!$this->sqliteOk || !isset($params['dbname'])) return false;
		if ($this->sqliteVers >= 3) {
			$this->sqlConn = new SQLite3($params['dbname']);
			//echo 'Connecting:';var_dump($this->sqlConn);
			return !!$this->sqlConn;
		} else {
			return false;
		}
	}
	
	// Run a simple SQL query
	protected function _sql($query) {
		if (!$this->sqliteOk || $this->sqliteVers < 3 || !$query || !$this->sqlConn) return;

		echo '<span style="font-size:7px;">'.$query.'</span><br>';
		
		$sqlR = $this->sqlConn->query($query);
		//var_dump($sqlR);var_dump($this->sqlConn);
		if ($sqlR === false) {
			$errcode  = isset($this->sqlConn->lastErrorCode) ? $this->sqlConn->lastErrorCode.' ' : '';
			$errmessg = isset($this->sqlConn->lastErrorMsg)  ? $this->sqlConn->lastErrorMsg : 'Unknown SQL error';
			$this->lastQueryError = (isset($this->sqlConn->lastErrorMsg)) ? $this->sqlConn->lastErrorMsg : 'SQL error';
			//echo $this->lastQueryError;
			return false;
		} else {
			return $sqlR;
		}
	}

	// auto primary key value for last insert operation
	protected function _sqlInsertID() {
		if (!$this->sqlConn) return 0;
		return $this->sqlConn->lastInsertRowID();
	}
	
	protected function _createSQLResultContainer() {
		return new OESQLite3Result(); //!! implement old sqlite??
	}

	// normally called BEFORE applying values to placeholders, to correct formats, ex for SQLite ` => " and " => '
	protected function _fixQueryFormat($query, &$rNeededExtraQueries) {
		/*
		No stored procedures
		" => ' 
		` => " or nothing?
		SQLite doesn't support FOREIGN KEY constraints
		SQLite: CREATE TABLE Blah (foreignId Integer REFERENCES OtherTable (id)); 
		
		PKeys:
		MySQL: int(11) NOT NULL AUTO_INCREMENT), ..., PRIMARY_KEY()
		SQLite: id INTEGER PRIMARY KEY ASC, ...
		
		CREATE TABLE IF NOT EXISTS `oe_logv0_user` 
			(oeid int(11) NOT NULL AUTO_INCREMENT, name TEXT NOT NULL, pw TEXT NOT NULL, email TEXT NOT NULL, 
			address TEXT NOT NULL, info TEXT NOT NULL, permissions int(11) NOT NULL, 
			PRIMARY KEY(oeid) [, FOREIGN KEY(key) REFERENCES table(column)]) ENGINE=InnoDB DEFAULT CHARSET=utf8		
		
		*/
		if (!$query) return $query;
		
		// 
		//$query = str_replace('"', "'", $query);
		//$query = str_replace('`', '"', $query); // important

		// types:
		$query = preg_replace('/int\(\d+\)/i', 'INTEGER', $query);
		$query = str_ireplace('double', 'REAL', $query);		
		
		// CREATE queries:
		if (isset($query[5]) && strtoupper(substr($query, 0, 6)) == 'CREATE') {
			
			// find table name, considering it's between first two `
			$posFirstQuote = strpos($query, '`');
			$posSecondQuote = strpos($query, '`', $posFirstQuote+1);
			$tableName = (!$posFirstQuote || !$posSecondQuote) ? null : substr($query, $posFirstQuote+1, $posSecondQuote-$posFirstQuote-1);
			//echo $tableName.'<br>';
		
			$query = str_ireplace(' NOT NULL AUTO_INCREMENT', ' PRIMARY KEY ASC', $query); // change primary key column format
			
			$query = preg_replace('/, PRIMARY KEY\(.+?\)/', '', $query); // remove ", PRIMARY KEY(column)" (lazy +)
			$query = preg_replace('/, FOREIGN KEY\(.+?\) REFERENCES .+?\(.+?\)/', '', $query); // remove ", FOREIGN KEY(key) REFERENCES table(column)" (lazy +)
			
			$query = str_ireplace(' ENGINE=InnoDB', '', $query);
			$query = str_ireplace(' DEFAULT CHARSET=utf8', '', $query);
			
			// move indexes to separate queries - sqlite doesn't support indexes in CREATE TABLE:
			if ($tableName) {
				$posRemoveStart = -1; $posRemoveEnd = -2;

				$query = str_replace('`(255))', '`)', $query); //!!check improve // remove index length: INDEX `ind_name` (`ind_text_column`(255)) => INDEX `ind_name` (`ind_text_column`)
				$posINDEX = -1;
				while (($posINDEX = strpos(strtoupper($query), ', INDEX `', $posINDEX+1)) !== false) { // find all index definitions, and move them into a seprate query:
					//echo $posINDEX;
					$posBRACKET  = strpos($query, '(', $posINDEX);
					$posINDEXend = strpos($query, ')', $posINDEX); 
					if (!$posBRACKET || !$posINDEXend) break;
					$codeINDEX1 = substr($query, $posINDEX+2, $posBRACKET-1-($posINDEX+2)+1); // ex INDEX `ind_table1_column1` 
					$codeINDEX2 = substr($query, $posBRACKET, $posINDEXend-$posBRACKET+1); // ex (`column1` ASC)
					$codeINDEX  = 'CREATE '.$codeINDEX1.' ON `'.$tableName.'` '.$codeINDEX2; // ex CREATE INDEX `ind_table1_column1` ON `table1` (`column1` ASC)
					//echo '<br>Added query: '.$codeINDEX.'<br/>';
					if (!stripos($codeINDEX, 'IF NOT EXISTS')) $codeINDEX = str_ireplace('INDEX ', 'INDEX IF NOT EXISTS ', $codeINDEX); // add only nonexisting indexes
					//$codeINDEX = str_replace('` (', '`(', $codeINDEX);
					//$codeINDEX = str_replace('`', '', $codeINDEX);
					$rNeededExtraQueries[] = $codeINDEX;
					if ($posRemoveStart < 0) 			$posRemoveStart = $posINDEX;
					if ($posRemoveEnd < $posINDEXend) 	$posRemoveEnd   = $posINDEXend;
					//echo '<br>Added query: '.$codeINDEX.'<br/>';
				}

				// remove INDEX code from original query:
				if ($posRemoveEnd > $posRemoveStart) $query = substr($query, 0, $posRemoveStart).substr($query, $posRemoveEnd+1);
			}
			 
		}
		
		//$query = 'SELECT OE_Table.test AS OE_Table.test, OE_Table2.score AS OE_Table2.score, OE_Table.OEID AS OE_Table.OEID, OE_Table2.OEID AS OE_Table2.OEID FROM OE_Table, OE_Table2 WHERE (OE_Table2.auto=OE_Table.OEID) AND (OE_Table2.score > 0) LIMIT 0,500';
		return $query;
	}
	
	protected function _sqlEscapeString($val) {
		if (!$this->sqliteOk) return $val;
		if ($this->sqliteVers < 3) return sqlite_escape_string($val);
		// SQLite3:
		if (!$this->_checkInit()) return $val;
		return $this->sqlConn->escapeString($val);
	}
	
	
}

